function [lambda,k,Jnew,znew,x]=armijo(x0,u,t,drct_u,drct_x0,grad,ki,ialg,...
	                               I_a,I_i,BL,BU,fit,free_x0,J)

%  [lambda,k,Jnew,znew,x] = armijo(x0,u,t,drct_u,drct_x0,grad,ki,ialg,
%                                  I_a,I_i,BL,BU,fit,free_x0,{J})
%
%  The global variable ALPHA, BETA, GOLDSTEIN_TEST and Minv 
%  must be set on entry.
%
%  INPUTS:
%  =======
%  ki(1) is initial k
%  ki(2) is lower bound on k
%  ki(3) is upper bound on k
%  ialg is the integration algorithm being used.
%  T is the transformation matrix if operating in transformed coordinates.
%  I_a is active index set ([] if no active set).
%  I_i is inactive index set ([] if no active set).
%  BL is matrix of lower bounds.
%  BU is matrix of upper bounds.
%  If no bounds, pass NULL.
%  J is the initial cost (optional)
%  {fit} means fit a parabola to starting point and the last point and 
%        accept minimizer as new steplength if it provides a
%        decrease in the cost function.
%	 Requires one extra simulation (fit will not be performed if 
%        GOLDSTEIN is set).
%
%  OUTPUTS:
%  ========
%  lambda --- stepsize.
%  k      --- lambda = BETA^k
%  Jnew   --- Cost at new at new iterate (x + lambda*drct).
%  znew   --- value of decision variables at selected stepsize.
%  x      --- new iterate (x + lambda*drct) if last simulation performed
%             was taken at stepsize lambda.  Otherwise armijo will return
%             x = [] so that the caller knows to resimulate system.
%
%  This function returns a modified Armijo step size.  The "modification"
%  deals with the possibility of control bounds and some other rules.
%  If and when the first Armijo test is satisfied (i.e. the stepsize
%  is small enough), and if GOLDSTEIN_TEST = 1, then the Goldstein
%  test is performed whenever the Armijo rule is satisfied for k=0.
%  This allows termination without taking a further trial step.
%  This is expecially nice for VM or Newton methods.
%  If GOLDSTEIN_TEST is set to 0, this test wont be performed.
%  NOTE: an alternative to using the GOLDSTEIN_TEST is to set ki(2) = 0
%  so that k=0 is automatically accepted if the armijo test is satisfied.
%  Also, if k <= 0, the steplength will not be allowed to increase
%  passed the point where a cost increase occurs.
%
%  If, during function evaluation, an IEEE_error occurs, armijo() will
%  automatically backtrack (by setting k = k+3).
% 
%  If the global variable ARMIJO_USER_TEST is defined,
%    then the stepsize will be choosen to guarantee that
%    ARMIJO_USER_TEST(j,x,x0,u,t,ialg,I_i,free_x0) <= 0.
%
%    I_i is a column vector indexing all elements of [u(:) ; x0(free_x0)]
%    that are not actively constrained.
%
%  This is useful for making sure that iterates stay in some set.
%
%  If the global variable "USER_FUNCTION_NAME" is defined,
%  pdmin() will call the m-file function specified by that variable
%  to evaluate the system functions and gradients.  This can be user to
%  construct a composite function from several functions.  For instance,
%  an exact penalty function can be formed thus making a constrained
%  problem look like an unconstrained problem.
%
%  Other possible modifications for Armijo:
%  1. In order to use alpha = 1/2 with Newton's method, on second half of
%     Armijo search, should allow k to continue decreasing until cost 
%     increases.  Actually, this may be a bad idea since once near a local
%     minimizer, k=0 will be accepted everytime if alpha < 1/2 (without an 
%     extra simulation if GOLDSTEIN_TEST = 1).  With this suggestion, 
%     you will have to do at least one extra simulation (probably two) to get
%     k=0 accepted.      
%  2. Add a check that if several iteration transpire then compute the
%     smallest distance necessary for a nonactive constraint to become
%     active and set the next stepsize to that.

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%

global ALPHA BETA GOLDSTEIN_TEST User_function_exist User_armijo_test_exist
global USER_ARMIJO_TEST_NAME USER_FUNCTION_NAME Minv DISP

if nargin < 15
  J=simulate(1,x0,u*Minv,t,ialg,1);
end

k=ki(1);
klower=ki(2);
kupper=ki(3);
NU = (1+ALPHA)/2;
ninputs = size(u,1);
length_u = size(u,2);
ncontrols = ninputs*length_u;
n_free_x0 = length(free_x0);

lambda = BETA^k;
drct = [drct_u(:) ; drct_x0(free_x0)];
slope = grad(I_i)'*drct(I_i);
adjust = 0;
%if (BL == [] & BU == []), bounds = 0; else bounds = 1; end
if ( isempty(BL) == 1) & (isempty(BU) == 1), bounds = 0; else bounds = 1; end

z = [u(:) ; x0(free_x0)];
if ~bounds
  ztry = z+lambda*drct;
else
  ztry = project(z,lambda*drct,BL,BU);
  %if I_a ~= []
  if (isempty(I_a) ==0)  
    adjust = grad(I_a)'*(z(I_a)-ztry(I_a));
  end,
end
utry = reshape(ztry(1:ncontrols),ninputs,length_u);
x0try = x0;
x0try(free_x0) = ztry(ncontrols+1:ncontrols+n_free_x0);

if User_function_exist
  [Jnew,x] = feval(USER_FUNCTION_NAME,x0try,utry*Minv,t,ialg,2);
else
  [Jnew,x] = simulate(1,x0try,utry*Minv,t,ialg,2);
end
delta = Jnew - J;

% First, satisfy Armijo criterion.

user_test = 0;
if User_armijo_test_exist
  user_test = ...
      feval(USER_ARMIJO_TEST_NAME,Jnew,x,x0try,utry*Minv,ialg,I_i,free_x0);
end

%-------------------D E C R E A S E   S T E P L E N G T H-------------------

Jold = J; zold = z;
if delta > ALPHA*(lambda*slope+adjust) | user_test > 0 | isnan(Jnew)

  while (delta > ALPHA*(lambda*slope+adjust) | user_test > 0 | isnan(Jnew)) ...
        & k < kupper
    if isnan(Jnew), 
      disp('Backtracking linesearch'); 
      k=k+3;
    else
      k=k+1;
    end
    lambda=lambda*BETA;
    if ~bounds
      ztry = z+lambda*drct;
    else
      ztry = project(z,lambda*drct,BL,BU);
      %if I_a ~= []
      if (isempty(I_a) ==0) 
        adjust = grad(I_a)'*(z(I_a)-ztry(I_a));
      end,
    end,
    utry = reshape(ztry(1:ncontrols),ninputs,length_u);
    x0try(free_x0) = ztry(ncontrols+1:ncontrols+n_free_x0);
    Jold=Jnew;
    if User_function_exist
      [Jnew,x]=feval(USER_FUNCTION_NAME,x0try,utry*Minv,t,ialg,2);
    else
      [Jnew,x]=simulate(1,x0try,utry*Minv,t,ialg,2);
    end
    delta = Jnew - J;
    if User_armijo_test_exist
      user_test = ...	
          feval(USER_ARMIJO_TEST_NAME,Jnew,x,x0try,utry*Minv,ialg,I_i,free_x0);
    end,
  end
  znew = ztry;

else

%-------------------I N C R E A S E   S T E P L E N G T H-------------------

  if ~(GOLDSTEIN_TEST & k==0 & delta>NU*(lambda*slope+adjust) & user_test<=0 ) 
    while delta <= ALPHA*(lambda*slope+adjust) & user_test <= 0 & ...
          k > klower & ~isnan(Jnew)
      k=k-1;
      lambda=lambda/BETA;
      zold = ztry;
      if ~bounds
        ztry = z+lambda*drct;
      else 
        ztry = project(z,lambda*drct,BL,BU);
        %if I_a ~= []
           if (isempty(I_a) ==0) 
          adjust = grad(I_a)'*(z(I_a)-ztry(I_a));
        end,
      end,
      utry = reshape(ztry(1:ncontrols),ninputs,length_u);
      x0try(free_x0) = ztry(ncontrols+1:ncontrols+n_free_x0);
      Jold=Jnew;
      if User_function_exist
        [Jnew,x]=feval(USER_FUNCTION_NAME,x0try,utry*Minv,t,ialg,2);
      else
        [Jnew,x]=simulate(1,x0try,utry*Minv,t,ialg,2);
      end
      if k < 0 & Jnew > Jold, break, end
      delta = Jnew - J;
      if User_armijo_test_exist
        user_test = ...
          feval(USER_ARMIJO_TEST_NAME,Jnew,x,x0try,utry*Minv,ialg,I_i,free_x0);
      end,
    end
    k = k+1;
    lambda = lambda*BETA;
    znew = zold;
    Jnew = Jold;
    x = [];      % This indicates that the current simulation
  end            % is not at steplength lambda.

end

%-------------------------Q U A D R A T I C   F I T-------------------------

% Find minimum of a quadratic fit to the first and last points of
% the Armijo stepsize procedure using the gradient from the first point.
% Requires only one extra function evaluation.
if fit & ~GOLDSTEIN_TEST

  % Note: slope = grad_u(I_i)*drct_u does not include contribution from active
  % constrained subspace because the "active" variables, if not already bound 
  % in which case utry(i)-u(i) = 0, are likely to get clipped during the
  % quadratic fit.
  a2 = Jnew - J - lambda*slope;
  if a2 > sqrt(eps)
    a2 = a2/lambda^2;
    lambdanew = -slope/(2*a2);
    if ~bounds
      ztry = z+lambdanew*drct;
    else
      ztry = project(z,lambdanew*drct,BL,BU);
    end,
    % The following conditional ensures the the active constraint set
    % is identified in a finite number of iterations.
    if lambdanew < lambda
      ztry(I_i) = znew(I_i);
    end
    utry = reshape(ztry(1:ncontrols),ninputs,length_u);
    x0try(free_x0) = ztry(ncontrols+1:ncontrols+n_free_x0);
    if User_function_exist
      [Jfitted,x]=feval(USER_FUNCTION_NAME,x0try,utry*Minv,t,ialg,2);
    else
      [Jfitted,x]=simulate(1,x0try,utry*Minv,t,ialg,2);
    end
    if isnan(Jfitted), return, end
    if User_armijo_test_exist
      user_test = ...
         feval(USER_ARMIJO_TEST_NAME,Jnew,x,x0try,utry*Minv,ialg,I_i,free_x0);
    end
    if ( Jfitted <= Jnew & user_test <= 0 )
      improvement = 100*(Jnew-Jfitted)/(J-Jnew);
      Jnew = Jfitted;
      lambda = lambdanew;
      if 0 & DISP == 2 & improvement > 0
       fprintf(1,'Quadratic fit helped;  improvement = %.2g%%.\n',improvement),
      end
      znew = ztry;
    else
      x = [];      % This indicates that the current simulation
    end            % is not at steplength lambda.
  end
end
