function [u,x,f0,lambda,I_i] = ...
    aug_lagr(x0,u0,t,Umin,Umax,params,N_inner,N_outer,ialg,method,tol,DISP)

% [u,x,f,lambda,I_i] = 
%        aug_lagrng([x0,{fixed,{x0min,x0max}},u0,t,Umin,Umax,params,
%                   N_inner,N_outer,ialg,{method},{[tol1,tol2]},{DISP})
%
% This function use an augmented Lagrangian method to minimize a function
% subject to equality constraints.  See description in pdmin.m for most of
% these inputs.
%
% tol1 is optimality tolerance (default 10*sqrt(eps))
% tol2 is constraint tolerance (default eps^(1/3))
%
% DISP controls how much printout is produced by pdmin.m

METHOD = 1;

% If METHOD = 1, the multipliers are updated at the end of each pdmin
% iteration by solving a least-squares problem.  This approach requires
% a finite number of outer iterations to adjust the penalty parameter.  The
% main draw back is that, because the multipliers change on every pdmin
% iteration, conjugate gradient methods and quasi-Newton methods may
% not provide fast convergence (just like in SQP methods!).  The multipliers
% are updated by solving a linear least-squares problems.
% If METHOD = 2, the multipliers are updated at the end of each outer
% iteration (method of multipliers).  The update is quite simple but an
% infinite number of outer iterations is required for convergence.
% 
% The penalties are adjusted at the end of each outer loop (reasonable
% since a good penalty value will be set after a finite number of
% increases).  Currently, there is no provision for reducing penalties,
% but it would be good to add such a thing.

% This function requires the functions  a_lagrng_fnc.m  and 
% multiplier_update.m .

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%


global USER_FUNCTION_NAME USER_TEST_NAME INFO PENALTIES MULTIPLIERS
global opt_program
%%%opt_program = 'aug_lagrng';
opt_program = 'aug_lagr';

if nargin < 9,
  disp('USAGE: '),
  disp(['[u,x,f,lambda,I_i] = aug_lagrng([x0,{fixed,{x0min,x0max}}],u0,t,' ...
        'Umin,Umax,params,'])
  disp(['                                '...
        'N_inner,N_outer,ialg,{method},{[tol1,tol2]},{DISP})']),
  return,
end
if nargin < 12, DISP = 0; end
if nargin < 11, tol = []; end
if nargin < 10, method = []; end
%if DISP == [], DISP = 0; end
%if method == [], method = ''; end
%if tol == []
if isempty(DISP), DISP = 0; end
if isempty(method), method = ''; end
if isempty(tol);
  grad_TOL = 10*sqrt(eps);
  constr_TOL = eps^(1/3);
end
TAU1 = 10;                    % Penalty increase factor.
TAU2 = 2;                     % Penalty decrease factor (not used).
GAMMA = 1/4;                  % Factor in determining acceptability of
                              % the current penalties.

%%%%USER_FUNCTION_NAME = 'a_lagrng_fnc';
USER_FUNCTION_NAME = 'a_lagfnc';
%%%%if METHOD == 1, USER_TEST_NAME = 'multiplier_update'; end
if METHOD == 1, USER_TEST_NAME = 'm_update'; end

INFO = simulate(0);

n_inequality_constraints = sum(INFO(7:10));
if n_inequality_constraints > 0,
  disp(['This program ignores inequality constraints.']);
end

n_eq_cnstr = sum(INFO(11:12));
if n_eq_cnstr == 0
  error('This problem has no constraints.  Use pdmin.m instead.');
end

if size(x0,2) > 1
  fixed_x0 = x0(:,2);
  free_x0 = find(~fixed_x0);
% add by Chen 97.09.23
else
  free_x0 = 0;
end

%
% --- Set initial multipliers and penalties.  It may help to increase
% --- initial penalties if optimization diverges.
%
MULTIPLIERS   = zeros(1,n_eq_cnstr);
PENALTIES     = ones(1,n_eq_cnstr);
old_constraint_values = ones(1,n_eq_cnstr);
C = zeros(1,n_eq_cnstr);

K = [];
u = u0;

%---------------------M A I N   O U T E R   L O O P----------------------

if METHOD == 1, N_inner = [N_inner(1);sqrt(eps)*10^min(6,N_outer)]; end
for iter = 1:N_outer

% --- PERFORM INNER LOOP ITERATION.
  [u,x,J,inform,I_a,I_i,M] = ...
            pdmin(x0,u,t,Umin,Umax,params,N_inner,ialg,method,K,DISP);
  x0(:,1) = x(:,1);

  % Break out of loop if pdmin.m terminated for a bad reason.
  if inform(1) < 0, break, end
  % On good pdmin loop termination, set next stepsize and scale factor
  % to their ending value on last loop.  Also, tighten pdmin tolerance
  % by a factor of 10.
  K=inform(3:4); 
  Scale = K(2);
  if inform(1) == 5, 
    K = [];            % Rescale.
  else
    if METHOD == 1
      if N_inner(2) > sqrt(eps)
        N_inner(2) = N_inner(2)/10;
      end
    end
  end

% --- Adjust penalties.	

  % The penalty updating technique used here is discussed in "Contrained
  % Optimization and Lagrange Multiplier Methods," Bertsekas, pp. 123.
  % But we don't increase penalties if constraint tolerance is satisfied.
  % There are other methods (such as that used in Lancelot or CFSQP).
  for k = 1:n_eq_cnstr,	
    C(k) = simulate(2,1+k,1);
  end
%  temp = abs(C) <= GAMMA*abs(old_constraint_value) | abs(C) <= constr_TOL;
% an "s" is missing previously. Yangquan Chen added in 1997.09.23
  temp = abs(C) <= GAMMA*abs(old_constraint_values) | abs(C) <= constr_TOL;
  % For I1 indices, penalties are big enough, update multipliers.
  % For I2 indices, penalties are not big enough, increase them.
  I1 = find(temp); I2 = find(~temp);
  old_constraint_values = C;

  if ( METHOD == 2 ),
    MULTIPLIERS(I1) = MULTIPLIERS(I1) - PENALTIES(I1).*C(I1);
  end
  PENALTIES(I2) = TAU1*PENALTIES(I2);

% --- COMPUTE THE NORM OF THE GRADIENT OF JUST THE LAGRANGIAN.
%     (pdmin returns the gradient of the augmented lagrangian.)

  [f0,x,dLu,dLx0]=simulate(1,x0(:,1),u,t,ialg,3);
  for k = 1:n_eq_cnstr;
    [dgu,dgx0] = simulate(2,k+1,2);
    dLu = dLu - MULTIPLIERS(k)*dgu;
    dLx0 = dLx0 - MULTIPLIERS(k)*dgx0;
  end    
  dLuM = dLu*M;
  dLuM = dLuM(:); dLu = dLu(:);
%  dLz1  = [dLuM(:) ; dLx0(free_x0)];
%  dLz2  = [dLu(:)  ; dLx0(free_x0)];
% added by Chen YQ, 97.09.23
  if (free_x0 == 0)
  dLz1  = [dLuM(:)];
  dLz2  = [dLu(:) ];
  else
  dLz1  = [dLuM(:) ; dLx0(free_x0)];
  dLz2  = [dLu(:)  ; dLx0(free_x0)];
  end
% end of adding.
  norm_gradL = Scale*sqrt(dLz1(I_i)'*dLz2(I_i));

% --- PRINTOUT SOME RELEVANT INFORMATION.

  % Separate fprintf() statements so first part gets printed only once.
  fprintf('\nCompleted %-2d Outer loop iterations.',iter);
  fprintf('\nMultipliers          : '); fprintf('%g ',MULTIPLIERS');
  fprintf('\nPenalties            : '); fprintf('%g ',PENALTIES');
  fprintf('\nConstraint Violations: ')
  fprintf('%g ',C)
  fprintf('\n')
  fprintf('Norm of unconstrained portion of Lagrangian gradient = %g\n',...
           norm_gradL);

% --- TEST FOR CONVERGENCE.

  if norm_gradL < grad_TOL*(1+abs(f0)) & all( abs(C) < constr_TOL )
    fprintf('\n')
    fprintf('Objective Value      : %g \n',f0)
    disp('Normal termination of outer loop.')
    break;
  end
  if iter == N_outer
    fprintf('\n')
    fprintf('Objective Value      : %g \n',f0)
    disp('Completed maximum number of outer iterations.')
    break;
  end

end
lambda = MULTIPLIERS;

clear global USER_FUNCTION_NAME USER_TEST_NAME INFO PENALTIES MULTIPLIERS
