function max_error = chk_grad(i,j,k,x0,u,t,ialg,params,central,DISP)
% function max_error = check_grad(i,j,k,x0,u,t,ialg,params,{central},{DISP})
%
%  This function checks the gradient calculation for simulate() by
%  comparing it to a finite-difference in the direction of du(i).
%  If you just want to verify that you have correct code for the
%  derivatives of h(t,x,u) and l(t,x,u) you should use check_deriv.msc.
%
%  Currently, only works for single-input systems.
%
%  DESCRIPTION OF INPUTS:
%  i ---  Index of control that will be perturbed for the finite-difference
%         calculation.
%         If i = 0, gradients wrt to controls will not be checked.
%
%  j ---  Index of initial state vector that will be perturbed for the
%         finite-difference calculation.
%         If j = 0, gradients wrt to initial states will not be checked.
%
%  k ---  For each trajectory constraint, k indicates the time index
%         at which the trajectory constraints will be evaluated.
%         The value of k doesn't effect function or endpoint constraint
%         computations.
%
%  x0---  Nominal initial state vector.
%  u ---  Nominal control vector.
%         Note: It's not wise to check gradients near equilibrium points or
%               near nominal control values.
%  
%  t ---  Time vector.
%
%  ialg  --- Integration algorithm.
%
%  params--- System parameters.  Specify [] if there are no system parameters.
%
%  central-- If a non-zero value of central is specified, then the finite-
%            difference calculations will be two-sided rather than one-sided.
%
%  DISP  --- Minimal information display if DISP = 0 (default = 1).

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%

if nargin < 7,
  disp(['USAGE: max_error = ' ...
        'check_grad(i,j,k,x0,u,t,ialg,{params},{central},{DISP})']),
  return,
end

fprintf('============================================================================\n')
if nargin < 10, DISP = 1; end
%if DISP == [], DISP = 1; end
if isempty(DISP), DISP = 1; end
if nargin < 9, central=0; end
if isempty(central), central=0; end
if central, pert_size= eps^(1/3); else pert_size=10*eps^(1/2); end
fprintf('Using perturbation size of %g\n',pert_size);
if nargin < 8, params=[]; end

info=simulate(0,params);
nfunctions = info(6);
nend_pt_constraints = sum(info(9:12));
ntraj_constraints = sum(info(7:8));
total = nfunctions+ntraj_constraints+nend_pt_constraints;
nparams = info(3);
points = length(t);
N = points-1;

if nparams ~= length(params)
  error('Wrong number of paramters.')
end

% Comment out the line below because it depends on integration method.
%if i>points, error(['i has to be <= ' int2str(N) '.']); end

if j>info(1), error(['j has to be <= ' int2str(info(1)) '.']); end

k = max(1,k);
if k>points, error(['k has to be <= ' int2str(points) '.']); end

dT = t(2)-t(1); 

J = zeros(total,1);
du_i = zeros(total,1);
dz_j = zeros(total,1);

% First store the nominal values for each function, requires only
% one state simulation.  If central, only du_i and du_z are needed.
f0=simulate(1,x0,u,t,ialg,1);
for fnc = 1:total,
  if ( fnc <= nfunctions )
    offset = fnc;
  elseif ( fnc <= nfunctions+ntraj_constraints )
    offset = nfunctions + (fnc-nfunctions-1)*points + k;
  else
    offset = fnc + ntraj_constraints*N;
  end
  if (fnc>1), f0=simulate(2,offset,1); end
  [du,dz]=simulate(2,offset,2);
  J(fnc) = f0;
  if i > 0, du_i(fnc) = du(i); end
  if j > 0, dz_j(fnc) = dz(j); end
end

max_error = 0;
for fnc = 1:total,
  
  if ( fnc <= nfunctions )
    offset = fnc;
    if DISP
      fprintf(1,'\n                    Evaluating function %d.\n',fnc);
    end,
  elseif ( fnc <= nfunctions+ntraj_constraints )
    offset = nfunctions + (fnc-nfunctions-1)*points + k;
    if DISP
      fprintf(1,'\n                    Evaluating trajectory constraint %d.\n',fnc-nfunctions);
    end
  else
    offset = fnc + ntraj_constraints*N;
    if DISP
      fprintf(1,'\n                    Evaluating endpoint constraint %d.\n',fnc-nfunctions-ntraj_constraints);
    end
  end

  if ( i > 0 )
    pert = pert_size*max(1,abs(u(i)));
    u1 = u; u2 = u;
    % Keep u(i) away from zero (if not using central difference).
    if u(i) < 0, pert_u = -pert; else pert_u = pert; end
    u1(i) = u(i)+pert_u;
    u2(i) = u(i)-pert_u;

    simulate(1,x0,u1,t,ialg,0);
    J1 = simulate(2,offset,1);
    if central
      simulate(1,x0,u2,t,ialg,0);
      J2 = simulate(2,offset,1);
      diff_u = (J1 - J2)/(2*pert_u);
    else
      diff_u = (J1 - J(fnc))/pert_u;
    end
    error_u = diff_u - du_i(fnc);
    if DISP, fprintf('error_u  = %g\n',error_u); end,
    max_error = max([max_error max(abs(error_u))]);
  end

  if ( j > 0 )
    pert = pert_size*max(1,abs(x0(j)));
    x01 = x0; x02 = x0;
    if x0(j) < 0, pert_x0 = -pert; else pert_x0 = pert; end
    x01(j) = x0(j)+pert_x0;
    x02(j) = x0(j)-pert_x0;

    simulate(1,x01,u,t,ialg,0);
    J1 = simulate(2,offset,1);
    if central
      simulate(1,x02,u,t,ialg,0);
      J2 = simulate(2,offset,1)	;
      diff_z = (J1 - J2)/(2*pert_x0);
    else
      diff_z = (J1 - J(fnc))/pert_x0;
    end
    error_z = diff_z - dz_j(fnc);
    if DISP, fprintf('error_x0 = %g\n',error_z); end,
    max_error = max([max_error max(abs(error_z))]);
  end
 

 if DISP
   if ( i > 0 )
     if central == 0, pert_u = (pert_u)^(1/2); end
       fprintf(1,'Relative error in control gradient = %g%%\n',...
                 100*abs(error_u/(diff_u+.0001)))
       if abs(error_u) > max([1,abs(J(fnc)),abs(du_i(fnc))])*abs(pert_u)
         disp('Bad Gradient.')
       else
        if abs(error_u) < max([1,abs(J(fnc)),abs(du_i(fnc))])*abs(pert_u)^(3/2)
          disp('Gradient OK')
        else
          disp('Gradient Fuzzy')
        end
      end
    end

    if ( j > 0 )
      if central == 0, pert_x0 = (pert_x0)^(1/2); end
      fprintf(1,'\nRelative error in x0 gradient = %g%%\n',...
      100*abs(error_z/(diff_z+.0001)))
        if abs(error_z) > max([1,abs(J(fnc)),abs(dz_j(fnc))])*abs(pert_x0)
          disp('Bad Gradient.')
        else
         if abs(error_z)<max([1,abs(J(fnc)),abs(dz_j(fnc))])*abs(pert_x0)^(3/2)
           disp('Gradient OK')
         else
           disp('Gradient Fuzzy,')
         end
      end
    end

  end
end
fprintf('\nMaximum error reported is %g.\n',max_error)
fprintf('============================================================================\n')
