function [new_t,new_u,new_lambda,sum_lte]=...
               distribu(t,u,x,ialg,lambda,n_free_x0,strategy,FAC,new_K,norm)

% [new_t,new_u,new_lambda,sum_lte] = 
%         distribute(t,u,x,ialg,lambda,n_free_x0,strategy,FAC,new_K,norm)
%
% This function computes a new grid that equidistributes the local
% truncation error for a fixed stepsize Runge-Kutta method.  The
% errors are estimated from the most current system simulation.
%
% INPUTS:
%          t        --- old grid
%          u        --- current u (spline coefficients)
%          x        --- current x
%	   ialg     --- integration order
%          lambda   --- (lambda = [lambda,istate] or lambda = []).
%                       If given, lambda is output from RIOTS when using NPSOL.
%                       Column 1 is vector of old Lagrange multipliers.
%                       Column 2 is vector of old activity indicators.
%          n_free_x0--- (only needed if lambda given) Number of free
%                       initial conditions.
%          strategy --- 1: Use equidistribution of plte,
%                       2: John Betts' strategy plus a mechanism for
%                          adding mesh points at times at or near 
%                          active trajectory constraints.
%                       (Add 10 to get relative error control.)
%                       (default is 11)
%                       3: Double the mesh,
%                       4: Change order to new_K.
%          FAC      --- (for use with strategies 1,2,11,12)
%                       The number of intervals in the new mesh is
%                       chosen to achieve an integration of the
%                       current integration accuracy divided by
%                       FAC.  Specifying FAC = [] or FAC = 0 will
%                       cause new length to equal old length.
%          new_K    --- order of the new, interpolated spline
%                       (default = K).  Specifying new_K = 0 will
%                       cause new_K = K.
%          norm     --- 0:         use 2-norm on each interval.
%                       otherwise: use maximum norm on each interval
%                       (default is 0)
% 
% For problems with control bounds use K <= 2.
% 
% OUTPUT:
%          new_t    --- a new integration mesh with which the local
%                       truncation error of the previous mesh is minimized,
%          new_u    --- u interpolated onto new mesh.
%          sum_lte  --- conservative estimate of the global error of 
%                       current simulation for each state.  
%          new_lambda --- if lambda is given in input and nargout > 2, then 
%                         new_lambda = [new_lambda,new_istate].

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%

if nargin < 7 | nargin > 10
  disp(['Usage: [new_t,new_u,new_lambda,sum_lte] = '...
       'distribute(t,u,x,ialg,lambda,n_free_x0,strategy,{FAC}'...
       ',{new_K},{norm})'])
  return
end
%if lambda ~[]
if (isempty(lambda) == 0) 
  if size(lambda,2) ~= 2, error('Must have lambda = [lambda,istate]'); end
end
K = size(u,2) - length(t) + 2;     % Order of input splines.
N = length(t)-1;

if nargin < 10, norm=0; end
%if nargin < 9, new_K=K; elseif new_K==0 | new_K == [], new_K=K; end
if nargin < 9, new_K=K; elseif ( (new_K==0) | isempty(new_K) ), new_K=K; end

% if nargin < 8, FAC=[]; elseif FAC==0; FAC=[]; end
if nargin < 8,
        FAC=[];
elseif (isempty(FAC)==0)
        if (FAC==0);FAC=[];end;
end

if nargin < 7, strategy=3; end

rel_error_control = fix(strategy/10);
strategy = strategy - rel_error_control*10;
if ( strategy < 1 | strategy > 4 )
  error('Please choose strategy = 1, 2, 3, or 4 (+10).')
end


%-----------------DETERMINE NEW SEQUENCE OF BREAKPOINTS------------------

if nargout == 4 | strategy == 1 | strategy == 2
  simulate(1,x(:,1),u,t,ialg,0);      % In case -DDIRECT is being used.
  lte = abs(simulate(7));
  % make lte_i relative to size of x_i.
  if rel_error_control,
    nstates = size(x,1);
    lte(1:nstates,:) = lte(1:nstates,:)./(1 + abs(x(:,1:N)));
    lte(nstates+1,:) = lte(nstates+1,:)/(1+abs(simulate(2,1,1))/N);
  end
  sum_lte = sum(lte')';

  if size(lte,1) > 1, 
    if (norm == 0), 
      lte = sqrt(sum(lte.*lte)); 
    else 
      lte = max(lte); 
    end
  end
  % plte is an estimate of the principle local truncation error.
  plte = lte./((1-(.5^ialg))*(diff(t).^(ialg+1)));
  %error = plte.^(1/(ialg+1)).*diff(t);
  % error is a reserved word. Let's choose another name. By Jinsong
  error_p = plte.^(1/(ialg+1)).*diff(t);

  % Safeguard against error_i too small which means delta_i too small.
  % This also ensures that the mesh is quasi-uniform.
  max_error = max(error_p);
  error_p = max( error_p , max_error/50 );
  rel_error = error_p/mean(error_p);

  % The following provides an estimate of how much reduction in plte
  % a redistribution with the same size mesh will produce.
  redistribute_factor = ( max(error_p)/min(error_p) )^(1/ialg);
  fprintf('\nRedistribution factor = %g\n',redistribute_factor)
%  if redistribute_factor < 2 & (strategy == 1 | strategy == 2)
%    strategy = 3;	 
%  end
end

if strategy == 4,
  new_t = t;
  if new_K == K
    disp('Changing spline order.')
    new_u = u;
    new_lambda = lambda;
    return;
  end
elseif strategy == 3
  disp('Doubling mesh.')
  new_t = zeros(1,2*N+1);
  new_t(1:2:2*N+1) = t;
  new_t(2:2:2*N+1) = (t(1:N) + t(2:N+1))/2;
else
  disp('Redistributing mesh.')
  if max_error < eps,
    new_t=t;new_u=u;
    disp('No integration error!')
    return;
  end

  if strategy == 1,
    % Estimate the new discretization level required to achieve error of TOL.
%    if FAC == []
    if isempty(FAC);
      new_N = N;
    else
      new_N = max( ceil( N*( FAC / max(rel_error) )^(1/ialg) ) , N );
      fprintf('New mesh will contain %d intervals.  ',new_N);
      fprintf('Old mesh contained %d intervals. \n',N);
    end

    % We are going to redistribute the mesh so that
    % plte_j^(1/ialg+1)*(t_j+1 - t_j) is constant.
    % The rest of this is lifted from newknt.m

    distfn = fnint(ppmak(t,error_p));
    % The total variation of  distfn  is its value at  break(l+1) , i.e,
    var = fnval(distfn,t(N+1));
    if var==0 ,
      new_t=t(1) + [0:new_N]*(t(N+1)-t(1))/new_N;
    else
      %
      % Use the variation in distfn to compute a new mesh distribution.
      % This code is from newknt.m.  Includes check for flat regions
      % in case error(i) = 0 for some i's.
      %
      steps = [0:new_N]*(var/new_N);
      new_t=zeros(1,new_N+1);new_t([1 new_N+1])=t([1 N+1]);
      [junk,coefs]=ppbrk(distfn);
      % The next two lines prevent division by zero if some interval 
      % has no error.
      flats=find(coefs(:,1)==0);
      if (~isempty(flats)),coefs(flats,1)=coefs(flats,1)-1;end
      t=t-t(1);
      invbrk=[coefs(:,2)' var];
      coefs=[ones(N,1)./coefs(:,1),t(1:N)'];
      new_t(2:new_N)=new_t(1) + fnval(ppmak(invbrk,coefs,1),steps(2:new_N));
    end

  else  % strategy == 2.
    % Estimate the new discretization level required to achieve error of TOL.
%    if FAC == []
    if isempty(FAC);
      scale = 1;
    else
      scale = ( FAC / max(rel_error) )^(1/ialg);
    end
    rel_error = rel_error*scale;

    CUTOFF = 0.25;
    % Divide each interval into nt intervals where 
    % nt(i) = 0 if rel_error(i) < CUTOFF
    % nt(i) = 1 if 0.2 <= rel_error(i) < 1
    % nt(i) = round(rel_error(i)) otherwise.
    % If an interval is to be removed, add its relative error to the
    % relative error of the previous interval.
    % 
    % NEW: Before doing the above, set rel_error(i) = max(1,rel_error(i)+0.5) 
    % if a trajectory constraint is active at or near time t(i).

%    if lambda ~= []
    if (isempty(lambda)==0)
      ncontrols = size(u,1)*size(u,2);
      N2 = ncontrols + n_free_x0;
%     [lambda_le_ineq,lambda_le_eq,lambda_nle_ineq,lambda_nle_eq,...
%       lambda_lt,lambda_nlt,lambda_obj]=sort_lambda(lambda(:,1),N+1,N2);
      [lambda_le_ineq,lambda_le_eq,lambda_nle_ineq,lambda_nle_eq,...
       lambda_lt,lambda_nlt,lambda_obj]=sort_lmd(lambda(:,1),N+1,N2);


      traj_multipliers = [lambda_lt,lambda_nlt];

%      if traj_multipliers ~= []
      if (isempty(traj_multipliers) ==0 )
        % zero-out multlipiers that are negligible.
        non_active = find(abs(traj_multipliers) < sqrt(eps))';
        traj_multipliers(non_active) = zeros(length(non_active),1);

        % Put zeros() on top of traj_multipliers so that any() is
        % operating on a matrix and even if traj_multipliers only has
        % one row.  That way any() returns a row vector of 1's and 0's
        % which indicate times at which trajectory constraints are
        % active.
        active = any([zeros(length(traj_multipliers),1),traj_multipliers]');
	% Now add 1's to (just ) the left of any 1 in active.
        I = any([[active(2:51) 0];active]);
	% Now shorten I and edge to be the same length as rel_error.
	I = I(1:length(I)-1); 
    % The following line is commented out by Jinsong Liang since the variable edge
    % conflicts with a function name edge, which causes an erros in Matlab
    % R14 SP1; also I did not find this variable edge is used anywhere except here
    %edge = edge(1:length(edge)-1);
        rel_error(I) = 1+rel_error(I);
      end
    end  

    for i=length(rel_error):-1:2
      if rel_error(i) < CUTOFF
        rel_error(i-1) = rel_error(i-1) + rel_error(i);
        rel_error(i) = 0;
      elseif rel_error(i) < 1
        rel_error(i) = 1;
      end
    end
    nt = round(rel_error);
    % Don't allow first breakpoint to be dropped!
    if nt(1) == 0, nt(1) = 1; end

    dt = diff(t);
    I = find(nt);
    nt = nt(I);
    new_dt = dt(I)./nt;
    new_t = zeros(1,1+sum(nt));
    index = 1;
    for i = 1:length(I),
      I2 = 0:nt(i)-1;
      new_t(index+I2) = t(I(i))+I2*new_dt(i);
      index = index+nt(i);
    end
    new_t(index) = t(N+1);
    fprintf('New mesh contains %d intervals.  ',length(new_t)-1);
    fprintf('Old mesh contained %d intervals. \n',N);
  end
end


%-------------------INTERPOLATE CONTROL ONTO NEW MESH-------------------

m = size(u,1);
N = length(new_t)-1;
new_u = zeros(m,N+new_K-1);

% Set up interpolation mesh.
%dt1 = (new_t(2) - new_t(1)) / new_K;
%dt2 = (new_t(N+1) - new_t(N)) / new_K;
%
%if ( fix(new_K/2) == new_K/2 )
%  % new_K is even.
%  space = [1:1:new_K/2];
%  tau = [new_t(1)+dt1*space new_t(2:N) new_t(N)+dt2*space];
%else
%  % new_K is odd.
%  space = [1:1:(new_K+1)/2];
%  midpoints = (new_t(2:N-1)+new_t(3:N))/2;
%  tau = [new_t(1)+dt1*space midpoints new_t(N)+dt2*space];
%end

% Do the interpolation.
if K>1
  new_knots = augknt(new_t,new_K);
  tau = aveknt(new_knots,new_K);
else
  new_knots = new_t;
  tau = new_t(1:N);   % 1st order splines are continuous from the right.
end
sp = spapi( new_knots , tau , fnval(spmak(augknt(t,K),u),tau) );
[knots,coeff] = spbrk(sp);
new_u = coeff;


%
% If lambda given in input and opt_program = 'NPSOL', 
% extend [lambda,istate] to new_t 
%

global opt_program
if nargout > 2
%  if lambda ~= [] & strcmp(opt_program,'NPSOL')
  if (isempty(lambda)==0) & strcmp(opt_program,'NPSOL')
    nparam = size(u,1)*size(u,2)+n_free_x0;
    [lambda,state] = ...
              extend(lambda(:,1)',lambda(:,2)',nparam,n_free_x0,t,new_t,new_K);
    new_lambda = [lambda',state'];
  else
    new_lambda = [];
  end
end
