function [int_error,norm_gLa,J,G,X,Ii] = est_erro(x0,u,t,Tf,ialg,lambda,I_i)

% [int_error,norm_gLa,J,G,x,Ii] = est_errors(x0,u,t,Tf,ialg,{lambda},{I_i})
%
% x0 = [x0,{fixed_x0,{}}]  where
% fixed_x0 is a nstates x 1 vector with a 1 in the i-th row indicating 
% that x0(i) is fixed, and a 0 indicating that x0(i) is a free variable.
% It is important to use the value of x0 at the current solution if there
% are free initial conditions!
%
% Tf is the final-time scale factor.  Use Tf=1 for fixed final time problems.
%
% lambda is a vector of lagrange multipliers returned from one of the
% optimization programs.  The ordering of the multipliers is determined
% automatically for each program using the global variable opt_program.
%
% I_i is the index of active control and initial state bounds.
% It does not need to be supplied if lambda is two columns with the
% second column giving the state of of each constraint (as lambda is
% supplied from RIOTS).  The default for I_i is 
% I_i = 1:(size(u,1)*size(u,2)+length(fixed_x0)).  That is, all bounds
% are assumed inactive.
% 
% int_error is a column vector.  The i-th component is the sum of the
% local truncation errors for the i_th component of the state vector.
% It should be proportional to (1/N)^ialg.
%
% norm_gLa is proportional to || z star - z_N star || sub I.
% When there are control constraints, norm_gLa is estimated by assuming the
% the indices in I_i are a good estimate of the unconstrained portion of 
% z star (t_i).  Need Lagrangian to be two times Lipschitz
% continuously differentiable with respect to z = [u ; x0(free_x0)].
% Use  control_error.m  to get good estimate from the information provided
% by this function.
%
% J is J(u_N star).  If u_N star --> u star as N --> inf, u star is a local
% minimizer and N is "sufficiently large", then J is a bound on J star.
%
% G is sum of actual constraint violations.  Each trajectory constraint
% violation is weighted by the length of the time interval over which that
% violation occurs.
%
% X is the trajectory x(u_N)(t) produced using variable stepsize integration.
%
% Ii = [Iui ; Ix0i] where
% Iui  is a set of indices into t[] that enumerate time intervals in which
%      control constraints are inactive and,
% Ix0i is indices of x0(free_x0) that are unconstrained.

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%

if nargin < 4,
  disp(['Usage: [int_error,norm_gLa,J,G,x,Ii] = '...
        'est_errors(x0,u,t,Tf,ialg,lambda,{I_i})'])
  return
end

if nargin < 6, lambda = []; end
if size(x0,2) > 1, 
  fixed_x0 = x0(:,2); x0 = x0(:,1);
else 
  fixed_x0 = ones(size(x0)); 
end

%---------------------COMPUTE INTEGRATION ERROR------------------------

% Note: it is not true that cost_error - int_error is an estimate for
% the control error because the order of accuracy of trajectory
% solutions can be greater than the order of accuracy of control
% solutions.  This is the case even if the control representation has
% the same order as the integration routine!
% Thus, this quantity is likely to be an under-estimate of the control error.

if ialg >= 1 & ialg <= 4
  simulate(1,x0,u,t,ialg,0);
  int_error = sum(abs(simulate(7))')';
else
  int_error = zeros(size(x0,1)+1,1);
end

if nargout < 2, return, end

global opt_program

%----------------------------------------------------------------------

points   = length(t);
length_u = size(u,2);
ninputs  = size(u,1);
N = points-1;
K = length_u - points + 2;
d = size(x0,1);
free_x0 = find(~fixed_x0);
n_free_x0 = length(free_x0);
ncontrols = size(u,1)*size(u,2);

%
% --- DETERMINE SET OF INDICES CORRESPONDING TO TIME INTERVALS
% ---       DURING WHICH CONTROL BOUNDS ARE NOT ACTIVE. 
%
N2 = ncontrols+n_free_x0;
if size(lambda,2) == 2, 
  istate = lambda(:,2); 
  lambda = lambda(:,1); 
  I_i = find(istate(1:N2) == 0);
elseif nargin < 7
  if strcmp(opt_program,'CFSQP')
    I_i = find(lambda(1:N2) == 0);
  else
    I_i = [1:ncontrols + n_free_x0]';
  end
end

%Ii = fill_indices(I_i,ninputs,length_u);
Ii = fill_ind(I_i,ninputs,length_u);
if n_free_x0 > 0
  Ii = [Ii ; I_i(I_i > ncontrols)];
end

%----------------S O R T   O U T   M U L T I P L I E R S----------------

%[lambda_le_ineq,lambda_le_eq,lambda_nle_ineq,...
%lambda_nle_eq,lambda_lt,lambda_nlt,lambda_f]=sort_lambda(lambda(:,1),points,N2);

[lambda_le_ineq,lambda_le_eq,lambda_nle_ineq,...
lambda_nle_eq,lambda_lt,lambda_nlt,lambda_f]=sort_lmd(lambda(:,1),points,N2);

lambda = [lambda_nlt(:) ; lambda_lt(:) ; lambda_nle_ineq ; ...
          lambda_le_ineq ; lambda_nle_eq ; lambda_le_eq];
  
info = simulate(0);
nfunctions = info(6);
n_traj_constraints = sum(info(7:8))*points;
n_ineq_constraints = sum(info(9:10));
n_eq_constraints   = sum(info(11:12));
nconstraints = n_traj_constraints + n_ineq_constraints + n_eq_constraints;

%  if strcmp(opt_program,'aug_lagrng'), nconstraints = sum(info(11:12)); end
% revised by Chen Yangquan 970925
if strcmp(opt_program,'aug_lagr'), nconstraints = sum(info(11:12)); end

%--------COMPUTE CONTROL ERROR AS THE NORM OF THE LAGRANGIAN.----------

% Perhaps for this estimate I should be using the augmented Lagrangian.

IALG = 5;     % Change to ialg=6 if the user is not supplying derivatives.
DT   = diff(t);

tol = max(1e-12,min(1e-4,sum(int_error)/1000));
[J,X,Lu,Lz]=simulate(1,x0,u,t,[IALG;1;tol;tol],3);
if isnan(J),
  error('est_error: Failed to integrate system using LSODA.')
end
if lambda_f(1) ~=1, Lu=Lu*lambda_f(1); Lz=Lz*lambda_f(1); end
for i = 2:nfunctions
  J = max(J,simulate(2,i,1));
  [du,dz] = simulate(2,i,2);
  Lu = Lu + lambda_f(i)*du;
  Lz = Lz + lambda_f(i)*dz;
end
G = 0;
for i = 1:nconstraints
  if abs(lambda(i)) > sqrt(eps)
    g = simulate(2,nfunctions+i,1);
    if i <= n_ineq_constraints
      if i <= n_traj_constraints
        G = G + max(0,g)*DT(rem(i-1,points)+1);
      else
        G = G + max(0,g);
      end
    else
      G = G + abs(g);
    end
    [du,dz] = simulate(2,nfunctions+i,2);
    Lu = Lu + ( -lambda(i) + abs(lambda(i))*g )*du;
    Lz = Lz + ( -lambda(i) + abs(lambda(i))*g )*dz;
  end
end
% Here's how to compute the norm of the gradient in function space:
% Let g be the gradient in function space, i.e. g = V^(-1) (Lu * Minv).
% Then <g,g>_L2 = < Lu*Minv , Lu*Minv >_LN = < Lu , Lu*Minv >_{l2} .
% To help visualize this, note that, for piecewise constant functions,
% g(t_i) = Lu(i)*N , i = 1 ,..., N  and  M = 1/N*eye(N,N) .

% At this point, transform() has probably already been called
% if K>2.  Thus it will return the previously computed M. 

%M = transform(t(1)+cumsum([0 DT]*Tf),K);
M = transfor( t(1)+cumsum([0 DT]*Tf), K);

LuM = Lu/M;               % Read comments above about this.

gLM = [LuM(:) ; Lz(free_x0)]; 
gL  = [Lu(:)  ; Lz(free_x0)];
norm_gLa = sqrt(gLM(Ii)'*gL(Ii));
