% This function is an easy-to-use interface to simulate(), form 2, that
% allows the user to easily check the value of various system constraints
% and objective functions.
%
% INPUTS:
% =======
% type   --- 'obj'   : objective function
%            'ei'    : endpoint inequality constraint
%            'ee'    : endpoint equality constraint
%            'traj'  : trajectory constraint
%
% num    --- specifies which function specified by 'type' is to be evaluated.
% 
% index  --- for trajectory constraints, evaluates the trajectory constraint
%            given by 'num' at the time point given by 'index'.
%            If 'index' is specified as a range, the trajectory constraint
%            will be evaluated at every time, t(i), for i in that index
%            range.  In this case, only function values will be computed;
%            no gradients will be computed.
%
% OUTPUTS:
% ========
% Returns the function value, f, and, if outputs requested, the
% gradients, du and dx0, and the adjoint p.  Remember that a constraint
% is violated if and only if the functions value is greater than zero.

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%

function [f,du,dx0,p] = eval_fnc(type,num,index)

if nargin < 2,
  disp('USAGE:  eval_fnc(type,action,num,index)')
  return
end

info=simulate(0);
n_obj  = info(6);
n_traj = sum(info(7:8));
n_ei   = sum(info(9:10));
n_ee   = sum(info(11:12));
points = info(14);

N1 = n_obj;
N2 = N1 + n_traj*points;
N3 = N2 + n_ei;

if num < 1, error('Argument ''num'' cannot be less than 1.'), end

%
% --- Compute approriate 'offset' so that the call
% --- simulate(2,offset,1) evaluates the correct function.
%

if strcmp(type,'obj')
  if num > n_obj, error('There aren''t that many objective functions.'), end
  offset = num;
elseif strcmp(type,'ei')
  if num > n_ei
    error('There aren''t that many endpoint inequality constraints.')
  end
  offset = N2 + num;
elseif strcmp(type,'ee')
  if num > n_ee
    error('There aren''t that many endpoint equality constraints.')
  end
  offset = N3 + num;
elseif strcmp(type,'traj')
  if num > n_traj  
    error('There aren''t that many trajectory constraints.')
  end
  if min(index) < 1 | max(index) > points
    error('Argument ''index'' is out of range.')
  end
  offset = N1 + (num-1)*points + index;
else
  error_msg = sprintf('There is no type ''%s''.  Try ''obj'', ''ei'', ''ee'', or ''traj''',type);
  error(error_msg)
end

%
% --- OK, call simulate().
%

if length(offset) == 1
  f=simulate(2,offset,1);
  if nargout > 1
    [du,dx0,p] = simulate(2,offset,2);
  end
else
  if nargout > 1, 
    disp('Gradients not computed for vector ''index'' input.')
  end
  f = zeros(1,length(offset));
  for i = 1:length(offset)
    f(i)=simulate(2,offset(i),1);
  end
end
