function [new_lambda,new_istate]=extend(lambda,istate,N,n_free_x0,t,new_t,new_K)

% lambda and istate should be short-fat matrices.
%
% N is the old number of decision variables.
% points = length(new_t) is the number of breakpoints.
%
% I only extend multipliers corresponding to nonlinear constraints.
% The others I set to zero.
%
% This program assumes you are using NPSOL.

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%

disp(' ')
disp('Extending multpliers into higher dimensional space')

ct = cputime;
points = length(t);
info = simulate(0);
ninputs = info(2);
nfunctions = info(6);
if nfunctions>1, error('Only implemented for single function objective.'), end
nlt_ineq = info(7);
lt_ineq  = info(8);
nle_ineq = info(9);
le_ineq  = info(10);
nle_eq   = info(11);
le_eq    = info(12);
LTI = info(13);
if LTI == 0
  nlt_ineq = nlt_ineq+lt_ineq; lt_ineq = 0;
  nle_ineq = nle_ineq+le_ineq; le_ineq = 0;
  nle_eq   = nle_eq+le_eq;     le_eq   = 0;
end

% ------------------- EXTEND MULTIPLIERS FOR CONSTRAINTS ---------------------

%  The order of constraints is as follows:
%    1   bounds on u
%    2   bounds on free x0's
%    3   linear endpoint inequality           
%    4   linear trajectory inequality         
%    5   linear endpoint equality             
%    6   nonlinear endpoint inequality        
%    7   nonlinear trajectory inequality      
%    8   nonlinear endpoint equality          

N1 = N-n_free_x0;
N2 = N;
N3 = N2+le_ineq;
N4 = N3+lt_ineq*points;
N5 = N4+le_eq;
N6 = N5+nle_ineq;
N7 = N6+nlt_ineq*points;
N8 = N7+nle_eq;
length_u = N1/ninputs;
K = length_u - points + 2;
if nargin < 7, new_K = K; end

% Extend multpliers corresponding to bounds on the controls using 
% the same interpolation method as is used to extend the controls.
if new_K > 1
  new_knots = augknt(new_t,new_K);
  tau = aveknt(new_knots,new_K);
  NN  = length(tau);
else
  new_knots = new_t;
  NN  = length(new_t)-1;
  tau = new_t(1:NN);   % 1st order splines are continuous from the right.
end
new_l1 = zeros(ninputs,length(new_t)+new_K-2);
% When extending istate, must not average adjacent istate values,
% so treat istate as a 1st order spline.
istate1 = reshape(istate(1:N1),ninputs,length_u);
temp   = fnval( spmak(t,istate1(:,1:points-1)) , tau );
sp_s1  = spapi(new_knots,tau,temp);
new_s1 = fnval(sp_s1,tau(1:NN));

NN = length(new_t);
if n_free_x0 > 0
  lambda2 = lambda(N1+1:N2); 
  istate2 = istate(N1+1:N2);
else 
  lambda2 = []; 
  istate2 = []; 
end
if le_ineq > 0
  lambda3 = lambda(N2+1:N3); 
  istate3 = istate(N2+1:N3);
else 
  lambda3 = []; 
  istate3 = []; 
end
if lt_ineq > 0
  % Always extend multipliers corresponding to trajectory 
  % constraints using linear interpolation.              
  % Don't actually need to bother extending multipliers for linear constraints.
  new_l4  = zeros(lt_ineq,points);
  istate4 = reshape(istate(N3+1:N4),lt_ineq,points);
  temp    = fnval( spmak(t,istate4(:,1:points-1)) , new_t(1:NN-1) );
  sp_s4   = spapi(new_t,new_t(1:NN-1),temp);
  new_s4  = fnval(sp_s4,new_t);
else
  new_l4 = [];
  new_s4 = []; 
end
if le_eq > 0
  lambda5 = lambda(N4+1:N5);
  istate5 = istate(N4+1:N5);
else
  lambda5 = [];
  istate5 = []; 
end
if nle_ineq > 0
  lambda6 = lambda(N5+1:N6);
  istate6 = istate(N5+1:N6);
else
  lambda6 = [];
  istate6 = []; 
end
if nlt_ineq > 0
  % Always extend multipliers corresponding to trajectory 
  % constraints using linear interpolation.              
  lambda7 = reshape(lambda(N6+1:N7),nlt_ineq,points);
  sp_l7   = spapi(new_knots,tau,fnval(spmak(augknt(t,K),lambda7),tau));
  new_l7  = fnval(sp_l7,new_t);  
  istate7 = reshape(istate(N6+1:N7),nlt_ineq,points);
  temp    = fnval( spmak(t,istate7(:,1:points-1)) , new_t(1:NN-1) );
  sp_s7   = spapi(new_t,new_t(1:NN-1),temp);
  new_s7  = fnval(sp_s7,new_t);
else
  new_l7 = [];
  new_s7 = []; 
end
if nle_eq > 0
  lambda8 = lambda(N7+1:N8);
  istate8  = istate(N7+1:N8);
else
  lambda8 = [];
  istate8  = []; 
end

new_lambda = [new_l1(:)' lambda2 lambda3 new_l4(:)' lambda5 lambda6 ...
	      new_l7(:)' lambda8];
new_istate  = [new_s1(:)' istate2 istate3 new_s4(:)' istate5 istate6 ...
	      new_s7(:)' istate8];

fprintf('Time to extend multpliers = %g seconds.\n',cputime-ct)
