function [jstar,xstar,ustar,pstar]=optimal(tvec,zeta0,sysnum)

%  SYSTEM 1:  
%  This returns the optimal trajectories for system 1.  As given in
%  "Linear Optimal Control System", Kwakernakak and Sivan, p. 212
%
%  SYSTEM 3 and LQR:
%  The optimal trajectories and control for the two systems given in
%  "Rates of Convergence for Discrete Apporximations to Unconstrained
%   Control Prolems," ISAM J. Numer. Anal. vol. 13, no. 4, Sept. 1976,
%   pp. 449-472.
%
%   SYSTEM  5: (switch.c)
%   Solution is given on page 122 of Bryson & Ho, "Applied Optimal Control."
%
%   SYSTEM  6: (robot.c)
%   SYSTEM  7: (bang.c)

if nargin < 3,
  disp('Usage: [jstar,xstar,ustar,pstar]=optimal(tvec,zeta0,sysnum)'),
  disp('sysnum        System')
  disp('------        ------')
  disp(' 1-3            1-3 ')
  disp('  4             LQR ')
  disp('  5           switch')
  disp('  6            robot')
  disp('  7            bang ')
  return,
end

if (sysnum == 1)
  alpha = .5;
  kappa = 150;
  rho = 1000;
  p1 = 0;
  gamma = sqrt(alpha^2 + kappa^2/rho);
  
  exp1 = exp(-gamma);
  exp2 = exp(gamma);
  den1 = ((gamma-alpha)*exp1 + (gamma+alpha)*exp2)/(2*gamma);
  den2 = -kappa^2*(exp1-exp2)/(2*gamma*rho);
  den = den1+den2*p1;
  
  exp1=exp(gamma*(tvec-1));
  exp2 = exp(-gamma*(tvec-1));
  th1 = ((gamma-alpha)*exp1 + (gamma+alpha)*exp2)/(2*gamma);
  th3 = -(exp1-exp2)/(2*gamma);
  th2 = th3*kappa^2/rho;
  th4 = ((gamma+alpha)*exp1 + (gamma-alpha)*exp2)/(2*gamma);
  pstar = 2*(th3+th4*p1)*zeta0./den;
  zeta  = (th1+th2*p1)*zeta0./den;

  xstar = zeta;
  ustar = -kappa*pstar/rho/2;
  jstar = xstar(1,:)'*pstar(1,:)/2;
end

if (sysnum == 3)
  a = -2/exp(3);
  den = -exp(3*tvec/2)*(a-1);
  num = a*exp(3*tvec);
  k = (2+num)./(1-num);
  xstar = (1-num)./den;
  pstar = k.*xstar;
%  ustar = (2+num)./den;
  ustar = -pstar;
  jstar = (exp(3)+exp(6)-2)/((2+exp(3))^2);
%  jstar = xstar(1,:)'*pstar(1,:)/2;
end


% The formula in Hager's papers is incorrect, this is correct.
if (sysnum == 4)
  xstar = cosh(1-tvec)/cosh(1);
  k = tanh(1-tvec);
  pstar = k.*xstar;
  ustar = -(k+.5).*xstar;
  jstar = exp(1)^2*sinh(2)/((1+exp(1)^2)^2);
%  jstar = xstar(1,:)'*pstar(1,:)/2;
end

% The solution below is valid for 0 < L <= 1/6 and t = [0,1].
if (sysnum == 5)
  points = length(tvec);
  L = 1/9; 
  L3 = 3*L;
  I1 = find(tvec<=3*L);
  I2 = find((tvec>3*L) & (tvec<=1-3*L));
  I3 = find(tvec>1-3*L);
  t1 = tvec(I1);
  t2 = tvec(I2);
  t3 = tvec(I3);

  ustar=zeros(size(tvec));
  v=ustar; x=ustar; pv=ustar; px=ustar;

  tt1 = 1 - t1/L3;
  tt3 = 1 - (1-t3)/L3;
  
  ustar(I1) = -2*tt1/L3;
  ustar(I3) = -2*tt3/L3;
  
  x(I1) = L*(1 - tt1.^3);
  x(I2) = L*ones(size(t2));
  x(I3) = L*(1 - tt3.^3);

  v(I1) = tt1.^2;
  v(I3) = -tt3.^2;
  
  px(I1) = 2/(L3^2)*ones(size(I1));
  px(I2) = zeros(size(I2));
  px(I3) = -2/(L3^2)*ones(size(I3));

  pv(I1) = -ustar(I1);
  pv(I2) = zeros(size(I2));
  pv(I3) = -ustar(I3);

  xstar = [x;v];			
  pstar = [px;pv];

  jstar = 4/(9*L);
end

if (sysnum == 6)
  jstar = []; pstar = []; ustar=[];
  I1 = find(tvec <= 1);
  I2 = find(tvec >  1);
  t1 = tvec(I1);
  t2 = tvec(I2);
  xstar = [t1/2                0.5*ones(size(I2)) ;
           0.5*ones(size(I1))  0*t2 ;
	   0*t1                (t2-1)/2;
	   0*t1                0.5*ones(size(I2))];
end

if (sysnum == 7)
  jstar = []; pstar = [];
  I1 = find(tvec <= 20);
  I2 = find(tvec > 20);	
  t1 = tvec(I1);
  t2 = tvec(I2);
  xstar = [t1.^2/2  -t2.^2+60*t2-600 ;
           t1        60 - 2*t2];       
  ustar = [ones(size(I1)) -2*ones(size(I2))];
end

