%RIOTSdem3	Solving a singular optimal control problem.

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%

clear all
clc,clf,home,echo on,hold off

% We will use the Goddard rocket maximum ascent optimal control problem
% for this demonstration.  This problem is a "singular optimal control
% problem".  It is also a free final-time problem.
%
% >!copy goddard.mex simulate.mex 
%
pause                              % TOUCH ANY KEY TO CONTINUE.
clear simulate
!copy goddard.mexw64 simulate.mexw64


% There are two versions of the goddard problem.  Both include a state
% endpoint constraint.  Additionally, if FLAGS = 1, then a trajectory
% constraint on the dynamic pressure (which is a nonlinear function of 
% the state variables) will be included in the problem.  First we'll
% solve the problem without the trajectory constraint.

clear simulate
global FLAGS
FLAGS = [];

pause                              % TOUCH ANY KEY TO CONTINUE.

% Since this is a free final-time problem, we had to augment that
% state equations with an extra variable that represents the "duration
% scale factor", Tf .  If the nominal time interval is  [a,b] , then
% the optimal time interval is given by  [a,a+(b-a)*Tf] .  So, the
% duration of the optimal manuever is  (b-a)*Tf .  Refer to the
% section "Free Final Time Problems" in the RIOTS user's manual for
% further details.
% 
% The initial conditions for the goddard rocket states are [0 ; 1 ; 1]
% but we have to add an initial conditions for the augmented state.

N  = 100;
t  = [0:1/N:1];
x0 = [0 ; 1 ; 1 ; 0.1];
u0 = ones(1,N+1);

% By setting the initial condition of the augmented state to 0.1 and
% choosing  t = [0:1/N:1] , we are providing an inital guess for the
% optimal time interval of  [0,0.1] .
%
% We use second order splines because the goddard rocket problem has
% control bounds of  0 <= u(t) <= 3.5 .

pause                              % TOUCH ANY KEY TO CONTINUE.

% Now we will solve this problem with riots() using the following call:
%
% TOL = [1e-4,1e-4,0,1e4];
% [u,x,f]=riots([x0,fixed],u0,t,0,3.5,[],100,4,TOL);
%
% The first input is  [x0,fixed] .  Fixed is a vector of the same length
% as  x0  that tells riots() which, if any, inital conditions are free
% to be modified.  Only those initial conditions whose corresponding
% entry in fixed is 0 are allowed to be modified.  In our case, only
% the initial condition for the augmented state is free to vary.  So
% we set

fixed = [1 ; 1 ; 1 ; 0];

% The vector  TOL  contains the convergence criterion tolerances.
% The first number is the optimality tolerance, the second is the
% allowable level of nonlinear constraint violations, third number
% depends on the particular SQP algorithm used---we just use 0 to get
% the default value.  The last number is an upper limit on the 
% largest conceivable magnitude of the decision variable.

pause                              % TOUCH ANY KEY TO CONTINUE.

TOL = [1e-4,1e-4,0,1e4]; % original setting
[u,x,f]=riots([x0,fixed],u0,t,0,3.5,[],100,4,TOL);

pause                              % TOUCH ANY KEY TO CONTINUE.

% This solution takes a lot of iterations to achieve the specified tolerance
% because the Goddard problem is singular.
% 
% To view the optimal control on the optimal time interval we have to
% extract the optimal duration scale factor , Tf , from the augmented
% state solution.  Then we can use Tf to scale the time vector t .

Tf = x(4,1)            % Get the optimal final time scale factor.

sp_plot(t*Tf,u)
xlabel('Time'), ylabel('Optimal Control')
title('Goddard rocket problem without trajectory constraint.')
axis([0 0.2 -0.5 4]);

pause                              % TOUCH ANY KEY TO CONTINUE.

% This looks OK, but not great.  We should redistribute the mesh
% points to achieve a better solution.  This is done with distribute()
% in the fourth demo.  Also, the solution would be a little better
% if we had use tighter tolerances for the termination criteria.
% 
% This solution is divided into three distinct regions.  First the
% thrust is on full, then the thrust is not at a bound, then there is
% no fuel left to burn so the thrust is zero.

pause                              % TOUCH ANY KEY TO CONTINUE.

% We can see how the velocity and altitude solutions look.

subplot(2,1,1)
plot(t*Tf,x(1,:))             % x1 is velocity
ylabel('Velocity')
title('Optimal velocity and altitude for the Goddard problem')

subplot(2,1,2)
plot(t*Tf,x(2,:))             % x1 is altitude (starting from 1 earth radius).
ylabel('Altitude'), xlabel('Time')

pause                              % TOUCH ANY KEY TO CONTINUE.
clc,home

% Now let's add in the trajectory constraint on the dynamic pressure.

FLAGS = 1;
clear simulate

% When we solve the problem this time, we have to supply a system
% parameter which gives, Q_max , the maximum allowable dynamic
% pressure.  The addition of this trajectory constraint will slow down
% riots() significantly.  But the solution will be obtained in 22
% iterations.
% 
% Q_max = 10;
% [u,x,f]=riots([x0,fixed],u0,t,0,3.5,Q_max,100,2,TOL);

pause                              % TOUCH ANY KEY TO CONTINUE.

Q_max = 10;
[u,x,f]=riots([x0,fixed],u0,t,0,3.5,Q_max,100,4,TOL);

pause                              % TOUCH ANY KEY TO CONTINUE.
 
Tf = x(4,1)                        % Get final time.
subplot(1,1,1)
sp_plot(t*Tf,u)
axis([0 0.25 -0.5 4]);
xlabel('Time'), ylabel('Optimal Control')
title('Goddard rocket problem with trajectory constraint.')

pause                              % TOUCH ANY KEY TO CONTINUE.

% Notice that, in the middle region, the control solution exhibits some
% unexpected oscillations.  This oscillation occurs in our numerical
% solution due to the fact that the trajectory constraint is active
% during the portion of the optimal trajectory that is "singular".
%
% Let's compute the dynamic pressure to show that it is at its upper
% bound of  Q_MAX = 10  in this middle region.

Arho0 = 12400; beta = 500;
rho=Arho0*exp(beta.*(1-x(2,:)));
q=0.5*rho.*x(1,:).^2;               % q(t) is the dynamic pressure.

pause                              % TOUCH ANY KEY TO CONTINUE.

subplot(2,1,1)
sp_plot(t*Tf,u)
axis([0 0.21 -0.25 3.75]);
ylabel('Optimal Control')
title('Goddard rocket problem with trajectory constraint.')

subplot(2,1,2)
plot(t*Tf,q)
ylabel('Dynamic Pressure')
xlabel('Time')
axis([0 0.21 0 11]);

pause                              % TOUCH ANY KEY TO CONTINUE.

% We can try to get rid of this unwanted control oscillation by adding
% a small penalty on the variation of the control derivative to the 
% objective function.  We do this by adding 1e-4 to the riots() input
% which specifies the maximum number of iterations:
%
% [u2,x2,f2] = riots([x0,fixed],u0,t,0,3.5,Q_max,[100;1e-4],4,TOL);
%
% Here we have added a penalty of 1e-4 times the variation of the control
% derivative to the objective function.

pause                              % TOUCH ANY KEY TO CONTINUE.

[u2,x2,f2] = riots([x0,fixed],u0,t,0,3.5,Q_max,[100;1e-4],2,TOL);

pause                              % TOUCH ANY KEY TO CONTINUE.

% First, let see how much the value of the objective function changed
% when we add the penalty on the variation of the control derivative.

abs(f2 - f)/abs(f)                 % Percentage change in objective.

% Well, that's very small!  

% Now let's compare the new control solution with the previous control
% solution.

pause                              % TOUCH ANY KEY TO CONTINUE.

subplot(1,1,1)
plot(t*Tf,u2,t*Tf,u,'-.')
xlabel('Time'),ylabel('Optimal Control')
title('Goddard rocket problem with trajectory constraint.')
legend('With penalty','Without penalty')

pause                              % TOUCH ANY KEY TO CONTINUE.
clc,home

% We might, at this point, wish to verify that our constraints are
% still satisfied even if there were less integration error.  
% It is quite possible that the constraints are only appear to be
% satisfied because of the discretization error caused by using a
% fixed stepsize Runge-Kutta integration routine.
%
% We can make this check using the variable stepsize integration
% routine with a very small integration tolerance.
% 
% The input [5 ; 1 ; 1e-8 ; 1e-8] specifies that variable stepsize
% itegration should be used to simulate the system with relative and
% absolute tolerances of 1e-8.  The 1 specifies how many internal
% knots should be used during computation of gradients.  This applies
% only to the variable stepsize algorithm (see the section on
% computing gradients with LSODA in the user's manual for more
% details).

[j,x] = simulate(1,x2(:,1),u2,t,[5 ; 1 ; 1e-8 ; 1e-8],2);

pause                              % TOUCH ANY KEY TO CONTINUE.

% Now we can check, say, the endpoint constraint:

f_num = 1+(N+1)+1;       % f_num = 1 gives the objective value, then there
simulate(2,f_num,1)      % are 101 points were the trajectory constraint
                         % gets evaluated and finally, there's the endpoint
			 % constraint.

% That's quite close to zero.  

pause                              % TOUCH ANY KEY TO CONTINUE.

% Let's also recompute the trajectory constraint and plot it.

Arho0 = 12400; beta = 500;
rho=Arho0*exp(beta.*(1-x(2,:)));
q=0.5*rho.*x(1,:).^2;               % q(t) is the dynamic pressure.

legend off
plot(t*Tf,q,t*Tf,10*ones(length(t)),'.')
title('Dynamic pressure recomputed using variable stepsize integration.')
xlabel('Time')
axis([0 0.22 -1 12]);

pause                              % TOUCH ANY KEY TO CONTINUE.

% This looks fine.  Let's check it out numerically

max(q) - 10

% So the maximum violation is still very small.

pause                              % TOUCH ANY KEY TO CONTINUE.
 
