%RIOTSdem4	Mesh redistribution.

%
% Copyright (c) 1996 by Adam L. Schwartz
% All right reserved.  
%

clear all
clc,clf,home,echo on,hold off, warning off

% We will again use the Goddard rocket maximum ascent optimal
% control problem for this demonstration.
% >!copy goddard.mex simulate.mex 
%
pause                              % TOUCH ANY KEY TO CONTINUE.
clear simulate
!copy goddard.mexw64 simulate.mexw64

% In the previous demonstration, we solved the Goddard problem and
% were not completely satisfied with the resulting control solution.
% We will repeat that computation now although, to exaggerate things a
% bit, we will use a discretization level of  N=50  instead of  N=100 .

clear simulate
global FLAGS
FLAGS = [];                        % No trajectory constraint.

N  = 50;
t  = [0:1/N:1];
x0    = [0 ; 1 ; 1 ; 0.1];
fixed = [1 ; 1 ; 1 ; 0];
u0 = ones(1,N+1);

TOL = [1e-4,1e-4,0,1e4];

pause                              % TOUCH ANY KEY TO CONTINUE.

[u,x,f]=riots([x0,fixed],u0,t,0,3.5,[],100,4,TOL);

pause                              % TOUCH ANY KEY TO CONTINUE.

% Recall that the solution requires a lot of iterations to achieve the
% specified tolerance because the Goddard problem a singular control
% problem.
%
% Here's what the control solution looks like.

Tf = x(4,1);
sp_plot(t*Tf,u)
axis([0 0.2 -0.5 4])
xlabel('Time'),ylabel('Optimal Control')
title('Solution of Goddard problem on uniform mesh')

pause                              % TOUCH ANY KEY TO CONTINUE.

% We can try to improve this solution by (i) increasing the number of
% mesh points (increasing the discretization level), (ii)
% redistributing the mesh points, or (iii) doing both.
% 
% First let's just redistribute the mesh.  Mesh redistribution is
% accomplished by the program distribute().  This program has several
% strategies for improving the mesh.  Here we will just use strategy
% one.  Refer to the user's manual for the other possibilities.

pause                              % TOUCH ANY KEY TO CONTINUE.

%[new_t,new_u]=distribute(t,u,x,4,[],[],1);
[new_t,new_u]=distribu(t,u,x,4,[],[],1);

% The required inputs to distribute() are the time vector for the
% current mesh t, the control and state trajectories  u  and  x  from
% the previous optimization, the integration algorithm that will be
% used in the next optimization (RK4 in this case), and the last input
% calls for strategy 1 to be used to redistribute the mesh.
% 
% The outputs  new_t  and  new_u  are, respectively, the redistribute
% mesh and the control u interpolated onto this new mesh.

pause                              % TOUCH ANY KEY TO CONTINUE.

% Now we'll resolve the problem using the following call to riots():
%
% [u2,x2,f2]=riots([x(:,1),fixed],new_u,new_t,0,3.5,[],200,4);
% 
% The difference between this call to riots() and the previous call to
% riots() is that we are using the current solution as the starting
% point for resolving the Goddard problem on the new mesh  new_t .
% Notice that instead of using [x0,fixed] as the first input we use
% [x(:,1) , fixed] .  This is because one initial condition for this
% problem is a free variable.  Thus  x(:,1) , the first column of the
% trajectory solution from the previous riots() call represents the
% current solution for the initial conditions.  Also, we're now going
% to use the default tolerance levels of  [1e-6,1e-4,0,1e4]  instead of
% [1e-4,1e-4,0,1e4]  because, presumably are new starting point is
% already close to a solution.

pause                              % TOUCH ANY KEY TO CONTINUE.

[u2,x2,f2]=riots([x(:,1),fixed],new_u,new_t,0,3.5,[],200,4);

pause                              % TOUCH ANY KEY TO CONTINUE.

% By improving the mesh through redistribution, the discretized
% problem becomes a more accurate approximation of the continuous-time
% problem and hence becomes itself "more singular".  This, explains
% why more iterations are required to obtain a solution on the new
% mesh.
% 
% Let's compare the solution from the uniform mesh to the solution we
% have obtain on the new, non-uniform mesh.

pause                              % TOUCH ANY KEY TO CONTINUE.

Tf = x(4,1);
Tf2 = x2(4,1);
plot(new_t*Tf2,u2,'-',t*Tf,u,'-.')
axis([0 0.2 -0.5 4])
legend('redistributed mesh','uniform mesh')
ylabel('Optimal Control')
title('Solutions before and after mesh redistribution (N=50)')
hold on
plot(new_t*Tf2,-0.5*ones(length(new_t)),'+')
set(gca,'xtick',[])

% The tick marks at the bottom of the plot indicate where each point
% in the re-distributed mesh is located.

pause                              % TOUCH ANY KEY TO CONTINUE.
clc,home

% Now we'll not only redistribute the mesh, but we'll also request
% enough new mesh points so that the overall integration error is
% decreased by a factor of 50.  This means that, at the c u r r e n t
% s o l u t i o n, the global integration error will be reduced by
% approximately a factor of 50.  We'll assume that our starting point
% was the solution we obtained using the original, uniform mesh.

%[new_t,new_u]=distribute(t,u,x,4,[],[],1,50);
[new_t,new_u]=distribu(t,u,x,4,[],[],1,50);

% So we only had to double the mesh to get a 50-fold decrease in
% integration error.

pause                              % TOUCH ANY KEY TO CONTINUE.

% Before solving the Goddard problem on this new mesh, let's actually
% compute an estimate for the global integration error on the new and
% the old meshes.

pause                              % TOUCH ANY KEY TO CONTINUE.

simulate(1,x(:,1),u,t,4,0);
lte1 = simulate(7);
glob_error1 = sum(sum(abs(lte1)))  % Global integraion error for uniform mesh.

simulate(1,x(:,1),new_u,new_t,4,0);
lte2 = simulate(7);
glob_error2 = sum(sum(abs(lte2)))  % Global int. error for non-uniform mesh.

% Well, it actually looks like the global integration error was
% reduced by a factor of almost 250!  Oh well.
 
pause                              % TOUCH ANY KEY TO CONTINUE.

[u3,x3,f3]=riots([x(:,1),fixed],new_u,new_t,0,3.5,[],200,4);

pause                              % TOUCH ANY KEY TO CONTINUE.

% The message "Kuhn-Tucker conditions satisfied but sequence did not
% converge" means that some values of the control keep changing more
% than expected from one iteration to the next.  This continuese to
% happen even at the end when the current solution appear close to optimal
% in the sense that the objective function cannot be furthered decreased.
% The cause of this condition arrises from the singularity of this
% control problem.

pause                              % TOUCH ANY KEY TO CONTINUE.

% Here's what the solution now looks like.  We can still see a fairly
% substantial glitch in the solution.  This is the likely cause of the
% warning that the sequence has not converged.  The trajectory solutions
% are not effected by this glitch because it gets integrated out.

hold off
legend off
Tf3=x3(4,1);
sp_plot(new_t*Tf3,u3);
hold on
plot(new_t*Tf3,-0.5*ones(length(new_t)),'+')
axis([0 0.2 -0.5 4])
set(gca,'xtick',[])
ylabel('Optimal Control')
title('Solution of Goddard problem on non-uniform mesh')

% Not bad.  Still some glitches, but this is a singular control
% problem after all.

pause                              % TOUCH ANY KEY TO CONTINUE.

echo off
hold off

clc,home
resp = '';
%while findstr('yn',resp) == []
while isempty(findstr('yn',resp))
  resp = input('Would you like to continue with another example (y,n)? ','s');
  resp = lower(resp);
end
if resp == 'n', return, end

clc,clf,home,echo on,hold off

% We will now use the Bang optimal control problem for
% this demonstration.
% >!copy bang.mex simulate.mex 
%
pause                              % TOUCH ANY KEY TO CONTINUE.
clear simulate
!copy bang.mexw64 simulate.mexw64



% This problem has linear dynamics with three state variables and one
% input, control bounds and endpoint constraints and free final-time.
% The optimal control is a bang-bang solution.  We have augmented the
% dynamics with one additional state to transform the problem in to
% a fixed final time problem.  The augmented state represents the 
% duration scale factor.
% 
% We will first solve this problem on a uniform mesh with 20 intervals
% using linear splines and the second order Runge-Kutta integration
% routine.

pause                              % TOUCH ANY KEY TO CONTINUE.

% First, let's setup the intial conditions.

x0 = [0 1 0 0; 0 1 0 0;1 0 .1 10]

% The first column is the initial condition with the optimal duraction
% set to 1.  The zero in the second column will be used to tell
% riots() that the third initial condition is a free variable.  The
% third and fourth columns represent upper and lower bounds on the
% initial conditions that are free variables.

pause                              % TOUCH ANY KEY TO CONTINUE.

clear simulate
N  = 20;
u0 = zeros(1,N+2-1);
t  = [0:10/N:10];                  % Set up the initial time vector so that
                                   % the intial time interval is [0,10]*1,
				   % where 1 is the initial estimate of the
				   % optimal duration scale factor.

pause                              % TOUCH ANY KEY TO CONTINUE.

% Here's our call to riots():
%
% [u,x,f]=riots(x0,u0,t,-2,1,[],100,2);

pause                              % TOUCH ANY KEY TO CONTINUE.

[u,x,f]=riots(x0,u0,t,-2,1,[],100,2);

pause                              % TOUCH ANY KEY TO CONTINUE.

Tf = x(3,1)

% The solution for the time interval is t*Tf = [0,10]*Tf.  Thus, the 
% solution for the final time is 10*Tf = 29.9813.  The actual solution
% for the final time is 30.
% 
% If we look at the solution, we will see that it is not very close to
% being the bang-bang solution we expect.

sp_plot(t*Tf,u)
xlabel('Time'),ylabel('Optimal Control')
title('Solution of Bang problem on uniform mesh')
axis([0 30 -2.5 1.5])

pause                              % TOUCH ANY KEY TO CONTINUE.

% Now, let's redistribute the mesh to see if we can't improve the solution.

%[new_t,new_u] = distribute(t,u,x,2,[],[],1);
[new_t,new_u] = distribu(t,u,x,2,[],[],1);

pause                              % TOUCH ANY KEY TO CONTINUE.

% Now we will re-solve the problem with this call to riots():
%
% [u2,x2,f2]=riots(x0,new_u,new_t,-2,1,[],100,2);

[u2,x2,f2]=riots(x0,new_u,new_t,-2,1,[],100,2);

pause                              % TOUCH ANY KEY TO CONTINUE.

Tf2 = x2(3,1)			   % Get the optimal duration scale factor.

% This optimal duration scaling gives the correct final time: 10*Tf = 30.

pause                              % TOUCH ANY KEY TO CONTINUE.

% Let's compare solutions.  We will plot both solutions and include little
% circles to show the mesh points.  The second solution is clearly much
% better even though the redistributed mesh has the same number of intervals
% as the original mesh.

subplot(1,2,1)
plot(t*Tf,u,t*Tf,u,'o')
xlabel('Time'),ylabel('Optimal Control')
title('Solution on uniform mesh')
axis([0 30 -2.5 1.5])

subplot(1,2,2)
plot(new_t*Tf,u2,new_t*Tf,u2,'o')
xlabel('Time')
title('Solution on redistributed mesh')
axis([0 30 -2.5 1.5])

pause                              % TOUCH ANY KEY TO CONTINUE.
